import pytest

from jetfuelburn.utility.tests import approx_with_units
from jetfuelburn.statistics import usdot

def test_usdot():
    assert 2024 in usdot.available_years()
    assert 'B787-800 Dreamliner' in usdot.available_aircraft(2024)


# %%
from jetfuelburn import ureg
from jetfuelburn.statistics import myclimate
from jetfuelburn.reducedorder import seymour_etal
from importlib import resources
import plotly.graph_objects as go
import pandas as pd


def figure_benchmarking():
    """
    Creates a Plotly figure comparing (...)
    
    Returns:
        go.Figure: Plotly figure object with payload vs range plot
    """
    pass

range_points = list(range(0, 10001, 250))

results = []
for current_range in range_points:
    fuelburn = myclimate.calculate_fuel_consumption(
        acft='standard aircraft',
        x=current_range * ureg.km
    )
    # Assuming fuelburn is a pint Quantity, store its magnitude
    results.append({
        'range': current_range,
        'fuelburn': fuelburn.magnitude
    })
df_myclimate_standard = pd.DataFrame(results)

results = []
for current_range in range_points:
    fuelburn = seymour_etal.calculate_fuel_consumption(
        acft='A321',
        R=current_range * ureg.km
    )
    # Assuming fuelburn is a pint Quantity, store its magnitude
    results.append({
        'range': current_range,
        'fuelburn': fuelburn.magnitude
    })
df_seymour_a321 = pd.DataFrame(results)

results = []
for current_range in range_points:
    fuelburn = seymour_etal.calculate_fuel_consumption(
        acft='B722',
        R=current_range * ureg.km
    )
    # Assuming fuelburn is a pint Quantity, store its magnitude
    results.append({
        'range': current_range,
        'fuelburn': fuelburn.magnitude
    })
df_seymour_b727 = pd.DataFrame(results)

fig = go.Figure()

fig.add_trace(go.Scatter(
    x=df_myclimate_standard['range'],
    y=df_myclimate_standard['fuelburn'],
    mode='lines+markers',
    name='MyClimate "standard aircraft"',
    line=dict(color='blue', width=1),
    marker=dict(symbol='x', size=6)
))

fig.add_trace(go.Scatter(
    x=df_seymour_a321['range'],
    y=df_seymour_a321['fuelburn'],
    mode='lines+markers',
    name='Seymour et al. A321',
    line=dict(color='red', width=1),
    marker=dict(symbol='x', size=6)
))

fig.add_trace(go.Scatter(
    x=df_seymour_b727['range'],
    y=df_seymour_b727['fuelburn'],
    mode='lines+markers',
    name='Seymour et al. B727',
    line=dict(color='green', width=1),
    marker=dict(symbol='x', size=6)
))

fig.update_layout(
    xaxis_title='Range [km]',
    yaxis_title='Fuel Burn [kg]',
    font=dict(family='Arial', size=11),
    legend=dict(
        yanchor="top",
        y=0.99,
        xanchor="left",
        x=0.01
    ),
    showlegend=True,
    xaxis=dict(
        showgrid=True,
        gridcolor='lightgrey',
        range=[0, 5000],  # Set x-axis limits
        minor=dict(
            showgrid=True,
            gridcolor='rgba(211, 211, 211, 0.5)', # Lighter grey for minor grid
            gridwidth=0.5
        )
    ),
    yaxis=dict(
        showgrid=True,
        gridcolor='lightgrey',
        range=[0, 35000], # Set y-axis limits (e.g., 0 to max value + 10%)
        minor=dict(
            showgrid=True,
            gridcolor='rgba(211, 211, 211, 0.5)', # Lighter grey for minor grid
            gridwidth=0.5
        )
    ),
    plot_bgcolor='white',
    width=600,
    height=400
)


import pytest

from jetfuelburn.utility.tests import approx_with_units
from jetfuelburn.reducedorder import (
    yanto_etal,
    aim2015,
    seymour_etal,
    lee_etal,
    eea_emission_inventory_2009,
    myclimate
)
from .fixtures.reducedorder import (
    fixture_yanto_B739,
    fixture_aim2015_B787,
    fixture_seymour_B738,
    fixture_lee_B732,
    fixture_eea_A320,
    fixture_myclimate_standard
)


def test_yanto_etal_B739(fixture_yanto_B739):
    make_case, params = fixture_yanto_B739
    
    for r, pl in params:
        input_data, expected_output = make_case(r, pl)
        calculated_output = yanto_etal.calculate_fuel_consumption(
            acft=input_data['acft'],
            R=input_data['R'],
            PL=input_data['PL'],
        )
        assert approx_with_units(
            value_check=calculated_output,
            value_expected=expected_output.to('kg'),
            rel=0.075
        )

def test_aim2015(fixture_aim2015_B787):
    make_case, distances = fixture_aim2015_B787
    
    for d in distances:
        input_data, expected_output = make_case(d)
        calculated_output = aim2015.calculate_fuel_consumption(
            acft_size_class=input_data['acft_size_class'],
            D_climb=input_data['D_climb'],
            D_cruise=input_data['D_cruise'],
            D_descent=input_data['D_descent'],
            PL=input_data['PL']
        )
        assert approx_with_units(
            value_check=sum(calculated_output.values()),
            value_expected=expected_output.to('kg'),
            rel=0.075
        )

def test_seymour(fixture_seymour_B738):
    make_case, ranges = fixture_seymour_B738
    
    for R in ranges:
        input_data, expected_output = make_case(R)
        calculated_output = seymour_etal.calculate_fuel_consumption(
            acft=input_data['acft'],
            R=input_data['R'],
        )
        assert approx_with_units(
            value_check=calculated_output,
            value_expected=expected_output.to('kg'),
            rel=0.075
        )

def test_eea2009(fixture_eea_A320):
    make_case, ranges = fixture_eea_A320
    
    for R in ranges:
        input_data, expected_output = make_case(R)
        calculated_output = eea_emission_inventory_2009.calculate_fuel_consumption(
            acft=input_data['acft'],
            R=input_data['R'],
        )
        assert approx_with_units(
            value_check=calculated_output['mass_fuel_total'].to('kg'),
            value_expected=expected_output.to('kg'),
            rel=0.075
        )


def test_myclimate_standard(fixture_myclimate_standard):
    make_case, params = fixture_myclimate_standard

    for x in params:
        input_data, expected_output = make_case(x)
        calculated_output = myclimate.calculate_fuel_consumption(
            acft='standard aircraft',
            x=input_data['x'],
        )
        assert approx_with_units(
            value_check=calculated_output,
            value_expected=expected_output.to('kg'),
            rel=0.075
        )


"""
# Files

## US Department of Transportation (DOT)

### `L_AIRCRAFT_TYPE.csv`

- Source: https://www.transtats.bts.gov/Fields.asp?gnoyr_VQ=FIH
- Description: Air Carrier Summary: T2: U.S. Air Carrier TRAFFIC And Capacity Statistics by Aircraft Type, Field: Aircraft Type
"""


import pint
from jetfuelburn import ureg
import json
from importlib import resources


class usdot():
    """
    This class contains methods to access statistical data on aircraft fuel consumption
    reported by "large certified air carriers" to the US Department of Transport (US DOT).

    Data is based on averages of fuel consumption per passenger mile and per ton mile,
    reported by the US DOT in table T2, summarizing Form 41 Schedule T-100.

    Terminology
    -----
    - **Form 41**
    
        Form 41 is a report that the U.S. Department of Transportation (DOT) generates
        based on data which large certified air carriers are required to provide.

        The _Financial Report_ part of this form includes balance sheet, cash flow, employment, income statement, fuel cost and consumption,
        aircraft operating expenses, and operating expenses.
        The _Air Carrier Statistics_ part of this form includes data on
        passengers, freight and mail transported.
        It also includes aircraft type, service class, available capacity and seats, and aircraft hours ramp-to-ramp and airborne.

        The reporting requirements of Schedule T-100 of Form 41 are defined in federal law: 

        | Reglation | Scope |
        | --------- | ----- |
        | [14 CFR 291.45](https://www.ecfr.gov/current/title-14/section-291.45) | General |
        | [Appendix A to Subpart E of Part 291, Title 14](https://www.ecfr.gov/current/title-14/part-291/appendix-Appendix%20A%20to%20Subpart%20E%20of%20Part%20291) | US Air Carriers |
        | [Appendix A to Part 217, Title 14](https://www.ecfr.gov/current/title-14/part-217/appendix-Appendix A to Part 217) | Foreign Air Carriers |

    - **Schedule** (in the context of Form 41)

        A schedule is a specific section of the Form 41 that contains
        a particular type of data:

        > "the Air Carrier Financial Reports (Form 41 Financial Data) (...)
        > Each table in this database contains a different type of financial report or “schedule” (...)"

        [Durso (2007) "An Introduction to DOT Form 41 web resources for airline financial analysis"](https://rosap.ntl.bts.gov/view/dot/16264/dot_16264_DS1.pdf)

    - Schedule **T100**

        > "Form 41 Schedule T-100(f) provides flight stage data covering both passenger/cargo
        > and all cargo operations in scheduled and nonscheduled services.
        > The schedule is used to report all flights which serve points in the United States
        > or its territories as defined in this part."

        [Appendix A to Part 217, Title 14](https://www.ecfr.gov/current/title-14/part-217/appendix-Appendix A to Part 217)
    
    - (table) **T2**

        > This table summarizes the T-100 traffic data reported by U.S. air carriers. The quarterly summary (...)

    - (table) **T1**

        > This table summarizes the T-100 traffic data reported by U.S. air carriers. The monthly summary (...)
        
    Warning
    -------
    Form 41 Schedule T-100 data can be downloaded from the US Department of Transportation website.
    Unfortunately, there are no permalinks to the data files, and even regular URLs are
    generated dynamically based on some JavaScript magic.
    
    The best way to obtain the correct files is to search by name, according to the following hierarchy:

    ```
    Database Name: Air Carrier Summary Data (Form 41 and 298C Summary Data)
    T2: U.S. Air Carrier TRAFFIC And Capacity Statistics by Aircraft Type
    ```

    On the `T2: (...)` page, click on the "Download" button in the "Data Tools" sidebar (left).
    The download should be an archive, containing the following file:

    ```
    T_SCHEDULE_T2.csv
    ```

    Aircraft types in this file are labeled using integer codes. These codes are
    defined in a separate file, which must also be downloaded.
    On the `T2: (...)` page, click on the `AircraftType` Field Name entry
    and on the following page, click on the "Download Lookup Table" button.

    See Also
    --------
    Working links to the relevant files (05-2025):
    
    - [US DOT: BTS: Air Carrier Summary Data (Form 41 and 298C Summary Data)](https://www.transtats.bts.gov/Tables.asp?QO_VQ=EGD&QO)
    - [US DOT: BTS: Air Carrier Summary Data: T2 (U.S. Air Carrier Traffic And Capacity Statistics by Aircraft Type)](https://www.transtats.bts.gov/Fields.asp?gnoyr_VQ=FIH)

    Example
    --------
    ```pyodide install='jetfuelburn'
    import jetfuelburn
    from jetfuelburn import ureg
    from jetfuelburn.statistics import usdot
    usdot.available_years()
    usdot.available_aircraft(usdot.available_years()[0])
    usdot.calculate_fuel_consumption_per_weight(
        year=2024,
        acft='B787-800 Dreamliner',
        R=1000 * ureg.nmi,
        W=1000 * ureg.kg
    )
    ```
    """

    _years = [2013, 2018, 2024]
    _aircraft_data = {}
    for year in _years:
        with resources.open_text("jetfuelburn.data.USDOT", f"USDOT_data_{year}.json") as file:
            _aircraft_data[year] = json.load(file)

    @staticmethod
    def available_years() -> list[int]:
        """
        Returns a sorted list of available years included in the model.
        """
        return sorted(usdot._years)
    @staticmethod
    def available_aircraft(year: int) -> list[str]:
        """
        Given a year, returns a sorted list of available ICAO aircraft designators included in the model.
        """
        return sorted(usdot._aircraft_data[year].keys())
    
    @staticmethod
    @ureg.check(
        None,
        None,
        '[length]',
        '[mass]',
    )
    def calculate_fuel_consumption_per_weight(
        year: int,
        acft: str,
        R: float,
        W: float,
    ) -> dict:
        r"""
        Given an aircraft name, range and payload weight, calculates the fuel burned during flight.

        Data is based on averages of fuel consumption per weight-distance flown by specific aircraft types,
        in specific years, reported by "large certified air carriers" to the US Department of Transport (US DOT):

        $$
        f = \frac{F(t=\text{year})}{WD(t=\text{year})} * R * W
        $$

        where:

        | Symbol | Units   | Description                                                                 |
        | ------ | ------- | --------------------------------------------------------------------------- |
        | $f$    | [kg]    | Fuel burned during flight                                                   |
        | $F$    | [kg]    | Total fuel consumption for a specific aircraft type in a specific year      |
        | $WD$   | [kg*km] | Total weight-distance flown on a specific aircraft type in a specific year  |
        | $R$    | [km]    | Range of the flight                                                         |
        | $W$    | [kg]    | Payload weight of the flight                                                |

        Warnings
        --------
        Aircraft types here are given as full-length strings, e.g. "Boeing 737-800", instead of the ususal ICAO designator "B738".
        These are taken directly from the `L_AIRCRAFT_TYPE.csv` US DOT lookup table file.

        Parameters
        ----------
        year : int
            Year of the data to be used for the calculation.
        acft : str
            Aircraft type to be used for the calculation.
        R : pint.Quantity
            Range of the flight.
        W : pint.Quantity
            Payload weight of the flight.

        Returns
        -------
        pint.Quantity
            Fuel burned during flight.

        Raises
        ------
        ValueError
            If the range or weight is negative.  
            If the year is not available in the model.  
            If the aircraft type is not available in the model for the given year.
        """
        if R.magnitude < 0 or W.magnitude < 0:
            raise ValueError(f"Range and/or weight must not be negative.")
        else:
            R = R.to('km')
            W = W.to('kg')

        if year not in usdot._years:
            raise ValueError(f"No data available for year '{year}'.")
        if acft not in usdot._aircraft_data[year]:
            raise ValueError(f"US DOT Aircraft Designator '{acft}' not found in model data.")
        else:
            aircraft_data = usdot._aircraft_data[year][acft]

        fuelburn = (aircraft_data['Fuel/Revenue Weight Distance'] * ureg('1/km')) * R * W
        fuelburn = fuelburn.to('kg')

        return fuelburn
    

    @staticmethod
    @ureg.check(
        None,
        None,
        '[length]',
    )
    def calculate_fuel_consumption_per_seat(
        year: int,
        acft: str,
        R: float,
    ) -> dict:
        r"""
        Given an aircraft name and range, calculates the fuel burned during flight per passenger.

        Data is based on averages of fuel consumption per passenger-distance flown by specific aircraft types,
        in specific years, reported by "large certified air carriers" to the US Department of Transport (US DOT):

        $$
        f = \frac{F(t=\text{year})}{paxD(t=\text{year})} * R
        $$

        where:

        | Symbol | Units   | Description                                                                   |
        | ------ | ------- | ----------------------------------------------------------------------------- |
        | $f$    | [kg]    | Fuel burned during flight                                                     |
        | $F$    | [kg]    | Total fuel consumption for a specific aircraft type in a specific year        |
        | $paxD$ | [kg*km] | Total passenger-distance flown on a specific aircraft type in a specific year |
        | $R$    | [km]    | Range of the flight                                                           |

        Warnings
        --------
        Aircraft types here are given as full-length strings, e.g. "Boeing 737-800", instead of the ususal ICAO designator "B738".
        These are taken directly from the `L_AIRCRAFT_TYPE.csv` US DOT lookup table file.

        Parameters
        ----------
        year : int
            Year of the data to be used for the calculation.
        acft : str
            Aircraft type to be used for the calculation.
        R : pint.Quantity
            Range of the flight.

        Returns
        -------
        pint.Quantity
            Fuel burned per passenger during flight.

        Raises
        ------
        ValueError
            If the range is negative.  
            If the year is not available in the model.  
            If the aircraft type is not available in the model for the given year.
        """
        if R.magnitude < 0:
            raise ValueError(f"Range must not be negative.")
        else:
            R = R.to('km')

        if year not in usdot._years:
            raise ValueError(f"No data available for year '{year}'.")
        if acft not in usdot._aircraft_data[year]:
            raise ValueError(f"US DOT Aircraft Designator '{acft}' not found in model data.")
        else:
            aircraft_data = usdot._aircraft_data[year][acft]

        fuelburn = (aircraft_data['Fuel/Revenue Seat Distance'] * ureg('kg/km')) * R
        fuelburn = fuelburn.to('kg')
        return fuelburn


from jetfuelburn import ureg
from jetfuelburn.reducedorder import lee_etal
from importlib import resources
import plotly.graph_objects as go
import pandas as pd


def figure_lee2010():
    """
    Creates a Plotly figure comparing my implementation of Lee et al. (2010) with the original data
    from the paper. The figure shows the payload vs range for a B732 aircraft.

    References
    ----------
    Lee, H. T., & Chatterji, G. (2010, September).
    Closed-form takeoff weight estimation model for air transportation simulation.
    In _10th AIAA Aviation Technology, Integration, and Operations (ATIO) Conference_ (p. 9156).
    doi:[10.2514/6.2010-9156](https://doi.org/10.2514/6.2010-9156) (Figure 6, data `f_res=0.08`)
    
    Returns:
        go.Figure: Plotly figure object with payload vs range plot
    """
    with resources.open_text("jetfuelburn.data.Lee2010", "data_fig_6_fres_const.csv") as file:
        df_fres_const = pd.read_csv(
            filepath_or_buffer=file,
            header=0,
            index_col=None
        )

    input_data = {
        'acft': 'B732',
        'W_E': 265825 * ureg.N,
        'W_MPLD': 156476 * ureg.N,
        'W_MTO': 513422 * ureg.N,
        'W_MF': 142365 * ureg.N,
        'S': 91.09 * ureg.m ** 2,
        'C_D0': 0.0214,
        'C_D2': 0.0462,
        'c': (2.131E-4) / ureg.s,
        'h': 9144 * ureg.m,
        'V': 807.65 * ureg.kph,
    }

    results = []
    for range in df_fres_const['range']:
        dict_results = lee_etal.calculate_fuel_consumption(
            acft=input_data['acft'],
            W_E=input_data['W_E'],
            W_MPLD=input_data['W_MPLD'],
            W_MTO=input_data['W_MTO'],
            W_MF=input_data['W_MF'],
            S=input_data['S'],
            C_D0=input_data['C_D0'],
            C_D2=input_data['C_D2'],
            c=input_data['c'],
            h=input_data['h'],
            V=input_data['V'],
            d=range * ureg.nmi
        )
        results.append({
            'range': range,
            'payload': dict_results['mass_payload'].to('lbs').magnitude
        })

    df_results = pd.DataFrame(results)

    fig = go.Figure()

    fig.add_trace(go.Scatter(
        x=df_fres_const['range'],
        y=df_fres_const['payload'],
        mode='lines+markers',
        name='Fig. 6 in Lee et al. (f_res=0.08)',
        line=dict(color='red', width=1),
        marker=dict(symbol='circle', size=6)
    ))

    fig.add_trace(go.Scatter(
        x=df_results['range'],
        y=df_results['payload'],
        mode='lines+markers',
        name='My Implementation (f_res=0.08)',
        line=dict(color='black', width=1),
        marker=dict(symbol='x', size=6)
    ))

    fig.update_layout(
        title='Payload vs Range',
        xaxis_title='Range [nmi]',
        yaxis_title='Payload [lbs]',
        font=dict(family='Arial', size=11),
        legend=dict(
            yanchor="bottom",
            y=0.01,
            xanchor="left",
            x=0.01
        ),
        showlegend=True,
        xaxis=dict(showgrid=True, gridcolor='lightgrey'),
        yaxis=dict(showgrid=True, gridcolor='lightgrey'),
        plot_bgcolor='white',
        width=600,
        height=400
    )

    return fig